#!/usr/bin/env perl

# Copyright INRA/CNRS

# Sebastien.Carrere@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package iANT::Common;

use strict;
use warnings;
use Carp;
use IO::File;
use CGI;
use iANT::Cfg;
use File::Basename;


require 'UNIVERSAL.pl';

=head1 NAME

iANT::DB::Feature - Object to deal with Entry's features pooled into one XML file


=head1 SYNOPSIS

    use iANT;

=head1 DESCRIPTION

    iANT package contains a set of methods (Get, Set, Dump)used by the following modules:
    
    iANT::Entry
    iANT::Entry::Definition
    iANT::Entry::Link
    iANT::Entry::Sequence
    iANT::Feature
    iANT::Feature::Database
    iANT::Feature::Software
    iANT::Feature::Sequence
    iANT::Feature::Hit
    iANT::Feature::Hit::Property
    iANT::Feature::Hit::Raw
    iANT::Organism
    iANT::Pattern
    iANT::DB::Feature
    iANT::DB::Fasta

=cut

=pod

=head3 procedure Dump

 Title        : Dump
 Usage        : $o_object->Dump();
 Prerequisite : none
 Function     : Dump an object in xml format
 Returns      : none
 Args         : $target:-optional; if empty or undef, print on STDOUT
                -if defined, print in the specified file (unix-like mode is mandatory)
 Globals      : none

=cut

sub Dump
{
    my ($self, $target, $no_xsl) = (@_);
    my $print_file = 0;
    if ((defined $target) && ($target ne ''))
    {
        if ($target !~ /^>/)
        {
            croak("\n\nYou must define the opening file mode: \">\" for overwrite or \">>\" for append mode\n\n");
        }
        else
        {
            open(OUT, "$target") || confess("$!:$target");
            select OUT;
            $print_file = 1;
        }
    }

    use ParamParser;
    my $o_param;
    $o_param = &iANT::Cfg::Load($o_param);

    my $xsl_file = $o_param->Get('entry_display');
    my $top_element = 'entry';
    
    if (ref ($self) =~ /iant::db::entry/i)
    {
        $xsl_file = $o_param->Get('entrydb_display') ;
        $top_element = 'entry_db';
    }
    elsif (ref ($self) =~ /iant::db::feature/i)
    {
        $xsl_file = $o_param->Get('entryfeatures_display') ;
        $top_element = 'entry_features';
    }
    
	my $xml_xsl_lines="";
    my $xml_xsl_header="";
    my $o_twig_xml = XML::Twig->new();
    $o_twig_xml->set_pretty_print('indented');
    my $xml = $self->GetXml();
    $o_twig_xml->parse($xml);
    
    if (!defined $no_xsl && (-e $xsl_file) )
    {
        my $o_twig_xsl = XML::Twig->new();
        $o_twig_xsl->parsefile($xsl_file);
        $xml_xsl_header = <<END;
<?xml-stylesheet type="text/xml" href="#stylesheet"?>
<!DOCTYPE $top_element [
<!ATTLIST xsl:stylesheet
  id	ID	#REQUIRED>
]>
END
        $xml_xsl_header ="\n" . $xml_xsl_header .  "\n";
        my $xsl_root_elt = $o_twig_xsl->root;
    
        eval {$xsl_root_elt->paste( $o_twig_xml->root)};
    }
    
    
    print '<?xml version="1.0"?>', $xml_xsl_header, "\n", $o_twig_xml->sprint();

    
	if ($print_file)
    {
        select STDOUT;
        close OUT;
    } 
	
    return;

}
=pod
 
=head3 Getters and Setters

=item Get

    Function     : Generic Getter
    Args         : $attribute

=cut

sub Get
{
    my ($self, $attribute) = (@_);
    if ((defined $attribute) && ($attribute ne ''))
    {
        return $self->{'__' . $attribute} if (defined $self->{'__' . $attribute});
    }
    else
    {
        confess("Get method needs an attribute name\n");
    }
}

=pod

=item Set

    Function    : Generic Setter
    Args        : $attribute
                  $value

=cut

sub Set
{
    my ($self, $attribute, $value) = @_;
    if ((defined $attribute) && ($attribute ne ''))
    {
        $self->{'__' . $attribute} = $value if (exists $self->{'__' . $attribute});
    }
    else
    {
        confess("Set method needs an attribute name\n");
    }
    return;

}

=pod

=item Backup

    Function    : Clone original object

=cut

sub Backup
{
    my $self = shift();
    if (!defined $self->{__o_backup})
    {
        my $o_clone = UNIVERSAL::Clone($self);
        $self->{'__o_backup'} = $o_clone;
    }
    return;

}

=pod

=item Restore

    Function    : Restore original object

=cut

sub Restore
{
    my $self = shift();
    if (defined $self->{'__o_backup'} && (ref($self->{'__o_backup'}) eq ref($self)))
    {

        #my $class   = ref($self);
        #my $o_original = bless {%{$self->{'__o_backup'}}}, $class;
        #return $o_original;
        %{$self} = %{$self->{'__o_backup'}};
    }
    else
    {
        confess("No backup available\n");
    }

}

=pod

=item NumericTest

    Function    : Do a numeric test from a string
    Args        : $left value
                  $operator (>,>=,<,<=,=) : empty operator is considered as '='
			   $right value
    Returns     : boolean

=cut

sub NumericTest
{
    my ($left, $operator, $right) = (@_);

    my $test_result = 0;

    if ($operator eq '>')
    {
        $test_result = 1 if ($left > $right);
    }
    elsif ($operator eq '>=')
    {
        $test_result = 1 if ($left >= $right);
    }
    elsif ($operator eq '<')
    {
        $test_result = 1 if ($left < $right);
    }
    elsif ($operator eq '<=')
    {
        $test_result = 1 if ($left <= $right);
    }
    elsif (($operator eq '=') || ($operator eq ''))
    {
        $test_result = 1 if ($left == $right);
    }
    return $test_result;
}

=pod

=item IsMatchingSelectionParameter

    Function    : Scan an object (iANT::Entry, iANT::Feature, ...) and
                  test if it is matching parameters (element/attributes)
    Args        : $o_scanned_object: the object to be scanned
                  $rh_selection_parameters: a hash ref of selection parameters 
    Returns     : boolean

=cut

sub IsMatchingSelectionParameter
{
    my ($o_scanned_object, $rh_selection_parameters) = (@_);

    my $match_selection_parameters = 0;
    foreach my $selection_parameter (keys %{$rh_selection_parameters})
    {
        if (ref($rh_selection_parameters->{$selection_parameter}) =~ /hash/i)
        {
            my $o_parameter_element = $o_scanned_object->Get('o_' . $selection_parameter);
            if (defined $o_parameter_element && ($o_parameter_element ne ''))
            {
                foreach my $element_selection_parameter (keys %{$rh_selection_parameters->{$selection_parameter}})
                {
                    if ($rh_selection_parameters->{$selection_parameter}->{$element_selection_parameter} =~ /^([\<\>=]+)(\d+)$/)
                    {
                        if (&NumericTest($o_parameter_element->Get($element_selection_parameter), $1, $2))
                        {
                            $match_selection_parameters = 1;
                        }
                        else
                        {
                            $match_selection_parameters = 0;
                            last;
                        }
                    }

                    elsif ($o_parameter_element->Get($element_selection_parameter) =~
                           /^$rh_selection_parameters->{$selection_parameter}->{$element_selection_parameter}$/)
                    {
                        $match_selection_parameters = 1;
                    }
                    else
                    {
                        $match_selection_parameters = 0;
                        last;
                    }
                }
                if (!$match_selection_parameters)
                {
                    last;
                }
            }
            else
            {
                my $ra_o_parameter_element = $o_scanned_object->Get('ra_' . $selection_parameter . 's');
                if (defined $ra_o_parameter_element && (ref($ra_o_parameter_element) =~ /array/i))
                {
                    foreach my $o_parameter_element (@{$ra_o_parameter_element})
                    {
                        foreach
                          my $element_selection_parameter (keys %{$rh_selection_parameters->{$selection_parameter}})
                        {

                            if ($rh_selection_parameters->{$selection_parameter}->{$element_selection_parameter} =~
                                /^([\<\>=]+)(\d+)$/)
                            {
                                if (&NumericTest($o_parameter_element->Get($element_selection_parameter), $1, $2))
                                {
                                    $match_selection_parameters = 1;
                                }
                                else
                                {
                                    $match_selection_parameters = 0;
                                    last;
                                }
                            }
                            elsif ($o_parameter_element->Get($element_selection_parameter) =~
                                   /^$rh_selection_parameters->{$selection_parameter}->{$element_selection_parameter}$/)
                            {
                                $match_selection_parameters = 1;
                            }
                            else
                            {
                                $match_selection_parameters = 0;
                                last;
                            }
                        }
                        if (!$match_selection_parameters)
                        {
                            last;
                        }
                    }
                }
            }
        }
        else
        {
            if ($rh_selection_parameters->{$selection_parameter} =~ /^([\<\>=]+)(\d+)$/)
            {
                if (&NumericTest($o_scanned_object->Get($selection_parameter), $1, $2))
                {
                    $match_selection_parameters = 1;
                }
                else
                {
                    $match_selection_parameters = 0;
                    last;
                }
            }
            elsif ($o_scanned_object->Get($selection_parameter) =~ /^$rh_selection_parameters->{$selection_parameter}$/)
            {
                $match_selection_parameters = 1;
            }
            else
            {
                $match_selection_parameters = 0;
                last;
            }
        }
    }
    return $match_selection_parameters;
}

=pod

=item function MultiFastaFile2DBEntry

 Title        : MultiFastaFile2DBEntry
 Usage        : MultiFastaFile2DBEntry($fasta_file,$o_param)
 Prerequisite : fasta2entrydb defined in $o_param
 Function     : Converts a multifasta file into an iANT::DB::Entry object
                using iANT.aa_dna.FastaEntryDBConverter.pl program
 Returns      : $o_db_entry (iANT::DB::Entry)
 Args         : $fasta_file the multifasta file path
                $o_param ParamParser object
 Globals      : none
 Note	    : TO BE OPTIMIZED ...
 
=cut

sub MultiFastaFile2DBEntry
{
    my ($fasta_file, $o_param) = (@_);

    if (!$o_param->IsDefined('fasta2entrydb'))
    {
        confess("fasta2entrydb not defined ...\n");
        return 0;
    }
    else
    {
        my $fasta2entrydb_cmd =
          $o_param->Get('fasta2entrydb') . ' --sequence ' . $fasta_file ;
	$fasta2entrydb_cmd .= ' --type ' . $o_param->Get('type') if ($o_param->IsDefined('type'));

        print STDERR "\n" . time() . "\tLaunch $fasta2entrydb_cmd\n" if ($o_param->IsDefined('verbose'));
        my $entry_db_xml = `$fasta2entrydb_cmd`;
        print STDERR "\n" . time() . "\tBuild DB::Entry\n" if ($o_param->IsDefined('verbose'));
        my $o_db_entry = New iANT::DB::Entry($entry_db_xml);
        print STDERR "\n" . time() . "\tDB::Entry [OK]\n" if ($o_param->IsDefined('verbose'));

        return $o_db_entry;
    }
}

=pod

=item function WhateverToiANTEntryList

 Title        : WhateverToiANTEntryList
 Usage        : WhateverToiANTEntryList($input,$o_param)
 Function     : From every type of input (directory, fof,file - entry xml, db::entry xml, fasta -, accession)
			 try to build an array of iANT::Entry objects;
			 Test Sequence is: 
				1. directory 
				2. file
					a. Entry
					b. DB::Entry
					c. Fasta
					d. file of file (recursive call of this function in that case)
				3. accession
 Returns      : $ra_o_entry (reference to a list of iANT::Entry)
 Args         : $input
                $o_param ParamParser object
 Globals      : [conf] extension_iant_entry , extension_iant_db_entry, unix_find
 
=cut

sub WhateverToiANTEntryList
{
    my ($input, $o_param) = (@_);

    my @a_o_entry = ();
$o_param = &iANT::Cfg::Load($o_param);
    

    my $extension_iant_entry = $o_param->Get('extension_iant_entry');
    my $unix_find            = $o_param->Get('unix_find');    
    my $extension_iant_db_entry = $o_param->Get('extension_iant_db_entry');
        
    if (&IsDir($input))
    {

        my @a_entry_files        = `$unix_find $input -type f -name \\*.$extension_iant_entry`;
        chomp @a_entry_files;
        foreach my $entry_file (@a_entry_files)
        {
            my $o_entry = New iANT::Entry::Display($entry_file);
            eval {$o_entry->Set('xmlfile', $entry_file)};
            if ($@)
            {
                $o_entry = New iANT::Entry($entry_file);
#                print STDERR "ERROR - $@\n";
            }
            push(@a_o_entry, $o_entry);
        }

        my @a_db_entry_files        = `$unix_find $input -type f -name \\*.$extension_iant_db_entry`;
        chomp @a_db_entry_files;
        foreach my $db_entry_file (@a_db_entry_files)
        {
            my $o_db_entry = New iANT::DB::Entry($db_entry_file);
            foreach my $o_entry (@{$o_db_entry->GetEntries()})
            {
                $o_entry->Set('xmlfile', $db_entry_file);
                push(@a_o_entry, $o_entry);
            }
        }
    }
    elsif (-e $input)
    {

#SEB Inversion des tests DBEntry PUIS Entry
        $input = &SquizzTest($input);
        my $o_db_entry = undef;
        if ($o_db_entry = &IsiANTDBEntry($input))
        {
            # my $o_db_entry = New iANT::DB::Entry($input);
            foreach my $o_entry (@{$o_db_entry->GetEntries()})
            {
#                $o_entry->Set('xmlfile', $input);
                push(@a_o_entry, $o_entry);
            }

        }
        elsif (&IsiANTEntry($input))
        {
            my $o_entry = New iANT::Entry::Display($input);
            $o_entry->Set('xmlfile', $input);
            push(@a_o_entry, $o_entry);
        }
        elsif (&IsFasta($input))
        {
            my $o_db_entry = MultiFastaFile2DBEntry($input, $o_param);
            foreach my $o_entry (@{$o_db_entry->GetEntries()})
            {
                if (! defined $o_entry->Get('xmlfile') || ($o_entry->Get('xmlfile') eq ''))
                {
                    $o_entry->Set('xmlfile',  $input . '.' . $o_entry->Get('accession') . '.' . $extension_iant_entry);
                }
                push(@a_o_entry, $o_entry);
            }

        }
        elsif (@a_o_entry = @{&IsFof($input, $o_param, 'WhateverToiANTEntryList')})
        {
            1;
        }

    }
    else
    {
        if (my $entry_file = &IsAccession($input, $o_param))
        {
            my $o_entry = New iANT::Entry::Display($entry_file);
            $o_entry->Set('xmlfile', $entry_file);
            push(@a_o_entry, $o_entry);
        }
        else
        {
            confess("Unknown input >$input<");
        }
    }

    return \@a_o_entry;
}

sub SquizzTest
{
    my $input = shift;
 
    eval 'use File::Which qw(which)';
    if ($@)
    {
        return $input;
    }
    
    my $exe_path = which('squizz');
    if (!defined($exe_path) || ($exe_path eq ''))
    {
        return $input;
    }
    else
    {
        my $format = `($exe_path -S $input) 2>&1 1>/dev/null`;
        chomp $format;
        if ($format =~ /UNKNOWN/)
        {
            return $input;
        }
        elsif ($format !~ /FASTA/)
        {
            print STDERR "SQUIZZ REFORMAT\n";
            my $fasta = `$exe_path -c FASTA $input 2> /dev/null`;
            chomp $fasta;
            
            my $fh_fasta = new IO::File(">$input.fasta") or Carp::croak("ERROR: can not open file >" . $input . ".fasta<");
            if (defined $fh_fasta)
            {
                print $fh_fasta $fasta . "\n";
                $fh_fasta->close;
                return "$input.fasta";
            }
            else
            {
                return $input;
            }
        }
    }
    
    return $input;
}

=pod

=item function IsAccession

 Title        : IsAccession
 Usage        : IsAccession($input,$o_param)
 Function     : Scan project data_dir to find an entry file with $input as an accession
 Returns      : $entry_file (entry xml file) if  found
			 0 if not found
 Args         : $input
                $o_param ParamParser object
 Globals      : [conf] data_dir
 
=cut

sub IsAccession
{
    my ($input, $o_param) = (@_);
    if ($o_param->IsDefined('data_dir'))
    {
        if (my $entry_file = &__FindiANTEntry($o_param->Get('data_dir'), $input, $o_param))
        {
            return $entry_file;
        }
        else
        {
            die(">$input< entry file cannot be found in " . $o_param->Get("data_dir") . " directory");
            return 0;
        }
    }
    else
    {
        die("If >$input< is an accession code, you must define the >data_dir< parameter to find this entry");
        return 0;
    }

}

=pod

=item function __FindiANTEntry

 Title        : __FindiANTEntry
 Usage        : __FindiANTEntry($scanned_drectory,$accession,$o_param)
 Function     : Scan a directory to find an entry file with $accession as an accession
			 using unix find command.
			 First, look for an entry directory named $accession; then look into this
			 directory if an entry file exist and is a real iANT::Entry file.
 Returns      : $entry_file (entry xml file) if  found
			 0 if not found
 Args         : $scanned_directory
			 $accession
                $o_param ParamParser object
 Globals      : [conf] unix_find, extension_iant_entry
 
=cut

sub __FindiANTEntry
{
    my ($scanned_dir, $accession, $o_param) = (@_);

    my $unix_find = $o_param->Get('unix_find');

    my $cmd           = "$unix_find $scanned_dir -type d -follow -name $accession";
    my @a_directories = `$cmd`;
    chomp @a_directories;

    if ($#a_directories > 0)
    {
        confess("More than one directory matching with $accession:\n" . join("\n", @a_directories));
    }
    my $accession_directory = $a_directories[0];

    if (defined $accession_directory && ($accession_directory ne ''))
    {
        my $entry_file = "$accession_directory/$accession" . '.' . $o_param->Get('extension_iant_entry');
        if ((-e $entry_file) && &iANT::Common::IsiANTEntry($entry_file))
        {
            return $entry_file;
        }
        else
        {
            return 0;
        }
    }
    return 0;

}

=pod

=item function IsDir

 Title        : IsDir
 Usage        : IsDir($input)
 Function     : Test if $input is a directory
 Returns      : boolean
 Args         : $input
 Globals      : none
 
=cut

sub IsDir
{
    my $input = shift();
    if (-d $input)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

=pod

=item function IsFastaGeneric

 Title        : IsFastaGeneric
 Usage        : IsFastaGeneric($input,$letters)
 Function     : Test if $input is a fasta file (has a header and a sequence string) 
 Returns      : boolean
 Args         : $input
			 $letters (regular expression of allowed letters in seuence string)
 Globals      : none
 
=cut

sub IsFastaGeneric
{
    my $input   = shift();
    my $letters = shift();
    $letters = '\w+' if ((!defined $letters) || ($letters eq ''));

    
    my $o_file = new IO::File;
    if ($o_file->open("< $input"))
    {
    my $header   = 0;
        my $sequence = 0;
        while (my $line = <$o_file>)
        {
            
		chomp $line;
	    
        if ($line ne '')
            {
                if ($line =~ /^>\S+/)
                {
                    $header = 1;
			
                }
                elsif ($line =~ /^$letters$/i)
                {
                    $sequence = 1;
                }
                else
                {
                    return 0;
                }
            }
            
        }
        if ($header && $sequence)
        {
            return 1;
        }
        else
        {
            return 0;
        }
        $o_file->close;

    }
    else
    {
        return 0;
    }
}

=pod

=item function IsFasta

 Title        : IsFasta
 Usage        : IsFasta($input)
 Function     : Test if $input is a fasta file (has a header and a sequence string) 
 Returns      : boolean
 Args         : $input
 Globals      : none
 
=cut

sub IsFasta
{
    my $input = shift();
    my $result = &IsFastaGeneric($input, '[\w\s]+');
    return $result;
}

=pod

=item function IsFastaNucleic

 Title        : IsFastaNucleic
 Usage        : IsFastaNucleic($input)
 Function     : Test if $input is a nucleic fasta file (has a header and a sequence string
			 containing only ATCGNU letters) 
 Returns      : boolean
 Args         : $input
 Globals      : none
 
=cut

sub IsFastaNucleic
{
    my $input = shift();
    my $result = &IsFastaGeneric($input, '[atcgun]+');
    return $result;
}



=pod

=item function IsiANTFasta

 Title        : IsiANTFasta
 Usage        : IsiANTFasta($input,['len','locus_tag'])
 Function     : Test if $input is a iant formatted fasta file (has a header with minimal information )
 Returns      : boolean
 Args         : $input
                OPTIONNAL $ra_mandatory_keys reference to a list of mandatory keys to be found
                            if not defined set to ['len','acc']
 Globals      : none
 
=cut

sub IsiANTFasta
{
    my $query_file            = shift();
    my $ra_mandatory_keys = shift;
    $ra_mandatory_keys = ['len','acc'] if (! defined $ra_mandatory_keys);
    
    my $ra_query_fastaheaders = &__GetFastaHeaders($query_file);
    if (&__ControliANTFastaHeaderFormat($ra_query_fastaheaders,$ra_mandatory_keys))
    {
        return 1;
    }
    return 0;
}

sub __GetFastaHeaders
{
    my $fasta_file     = shift();
    my @a_fastaheaders = ();
    my $fh_fasta_file = new IO::File($fasta_file) || Carp::croak("ERROR $fasta_file $!");
    if (defined $fh_fasta_file)
    {
        while (my $line = <$fh_fasta_file>)
        {
            chomp $line;
            if ($line =~ /^>/)
            {
                push(@a_fastaheaders, $line);
            }
        }
        $fh_fasta_file->close;
    }
    if ($#a_fastaheaders == -1)
    {
        Carp::croak("ERROR : $fasta_file is not a FASTA file (No header found) -- FATAL");
    }
    return \@a_fastaheaders;
}

sub __ControliANTFastaHeaderFormat
{
    my $ra_fastaheaders = shift;
    my $ra_mandatory_keys = shift;
    
    foreach my $fastaheader (@{$ra_fastaheaders})
    {
        foreach my $mandatory_key (@{$ra_mandatory_keys})
        {
            if ($fastaheader !~ /\s+$mandatory_key=\S+\s*/)
            {
                return 0;
            }
        }
    }
    return 1;
}

sub ControlDBIndex
{
    
    my ($dbfile, $method, $o_param,$system_msg_flux) = (@_);

    my ($provider, $program) = ($method =~ /^(\w+)-(\S+)/i);

    my %h_program_dbtype = (
                            'megablast' => 'dna',
                            'blastn'  => 'dna',
                            'tblastn' => 'dna',
                            'tblastx' => 'dna',
                            'blastp'  => 'aa',
                            'blastx'  => 'aa'
                            );

    $provider = 'netblast' if ($o_param->IsDefined('netblast'));

    my $extractcmd = $o_param->Get($provider . '-blast_extract');


    $extractcmd .= ' -n ' . $dbfile . ' ' if (($provider eq 'wu') && ($h_program_dbtype{$program} eq 'dna'));
    $extractcmd .= ' -p ' . $dbfile . ' ' if (($provider eq 'wu') && ($h_program_dbtype{$program} eq 'aa'));
    $extractcmd .= ' -p F ' . ' -d ' . $dbfile . ' -s '
      if (($provider eq 'ncbi') && ($h_program_dbtype{$program} eq 'dna'));
    $extractcmd .= ' -p T ' . ' -d ' . $dbfile . ' -s '
      if (($provider eq 'ncbi') && ($h_program_dbtype{$program} eq 'aa'));
    $o_param->Set('extractcmd', $extractcmd);
    $o_param->Set('blast2features',$o_param->Get($provider . 'blast2features') );
    

    my @a_index_extensions = split(
                                   /,/,
                                   $o_param->Get(
                                               $provider . '-blast_' . $h_program_dbtype{$program} . '_index_extensions'
                                     )
                                     );

    my $dbdir            = dirname($dbfile);
    my $dbfile_shortname = basename($dbfile);
    opendir(DBDIR, $dbdir) || confess("$! : $dbdir // $dbfile_shortname // $dbfile");
    my @a_files = readdir(DBDIR);
    close(DBDIR);

    my $build_index = 0;
    if ($o_param->IsDefined('force_reformat_db'))
    {
        $build_index = 1;
    }
    else
    {
        foreach my $index_extension (@a_index_extensions)
        {
            $build_index = 1;
            foreach my $existing_file (@a_files)
            {
                if ($existing_file =~ /^$dbfile_shortname\S+$index_extension$/)
                {
                    $build_index = 0;
                    last;
                }
            }
        }
    }
    if ($build_index)
    {
        my $build_index_command = $o_param->Get($provider . '-blast_index');
        my $dbtype_param        = '';

        $build_index_command .= ' -n ' . $dbfile if (($provider eq 'wu') && ($h_program_dbtype{$program} eq 'dna'));
        $build_index_command .= ' -p ' . $dbfile if (($provider eq 'wu') && ($h_program_dbtype{$program} eq 'aa'));
        $build_index_command .= ' -p F ' . ' -i ' . $dbfile
          if (($provider eq 'ncbi') && ($h_program_dbtype{$program} eq 'dna'));
        $build_index_command .= ' -p T ' . ' -i ' . $dbfile
          if (($provider eq 'ncbi') && ($h_program_dbtype{$program} eq 'aa'));

        $build_index_command = "($build_index_command > /dev/null ) 2> /dev/null"if (!$o_param->IsDefined('verbose'));
        
        print STDERR "\nWarning\n\n\t$dbfile is now indexed\n\t$build_index_command\n"
          if ($o_param->IsDefined('verbose'));
     system($build_index_command);
    }

}

=pod

=item IsFof

 Function     	:   Test if an input is a file of files
 Args          : $input - tested data
                 $o_param - ParamParser object
		       $source_sub - the function to apply on a file
 Returns		: $ra_list: a list reference of what is returned by $source_sub
 Globals		: none

=cut




sub IsFof
{
    my ($input, $o_param, $source_sub) = (@_);
    my $ra_list = [];
    my @a_files = ();
    my $o_file  = new IO::File("$input");
    return 0 if (!defined $o_file);

    
    while (my $line = <$o_file>)
    {
        chomp $line;
        push(@a_files, $line);

    }
    $o_file->close;

    foreach my $file (@a_files)
    {
        if (-e $file && $file =~/\w/) 
        {
            print STDERR "$file\n";
            my $new_ra_list = eval("&$source_sub('$file','$o_param')");
            if ($@ ne '')
            {
                print STDERR "ERROR  - $@";
            }
            else
            {
                push(@{$ra_list}, @{$new_ra_list});
            }
        }
    }

    return $ra_list;
}

sub GetFileNamesFromFof
{
    my ($input) = (@_);
    my @a_files = ();
    my $o_file  = new IO::File("$input");
    return 0 if (! -e $input);

    while (my $line = <$o_file>)
    {
        chomp $line;
	return 0 if (! -e $line);
        push(@a_files, $line);

    }
    $o_file->close;
    return \@a_files;
}

=pod

=item function IsiANTEntry

 Title        : IsiANTEntry
 Usage        : IsiANTEntry($input)
 Function     : Test if $input is an Entry Xml file 
			 (test root node name and try to build a iANT::Entry object) 
 Returns      : boolean
 Args         : $input
 Globals      : none
 
=cut

sub IsiANTEntry
{
    my $input = shift();
    eval 'use iANT::Entry';
    return 0 if ($@);


    my $o_entry;
    my $tmp_file = "/tmp/$$" .$^T;
    if (! -e $input)
    {
    	
    	my $o_file = new IO::File(">$tmp_file");
	print $o_file $input;
	$o_file->close;
	$input = $tmp_file;
    
    }


    my $o_file = new IO::File($input);
    
    if (!defined $o_file)
    {
        unlink($tmp_file) if (-e $tmp_file);
        return 0;
    }
    while (my $line = <$o_file>)
    {
        chomp $line;
        if ($line =~ /^<\?.+\?>\s*$/)
        {
            next;
        }
        elsif ($line =~ /<entry_db\s/)
        {
            unlink($tmp_file) if (-e $tmp_file);
            return 0;
        }
        else
        {
            if ($line =~ /<entry\s/)
            {
                eval {$o_entry = New iANT::Entry::Display($input)};
                if ($@)
                {unlink($tmp_file) if (-e $tmp_file);
                    return 0;
                }
                else
                {unlink($tmp_file) if (-e $tmp_file);
                    return $o_entry;
                }
            }
            else
            {
#                return 0;#seb debug pour xsl inline
                next;#seb debug pour xsl inline
            }
        }
    }
    $o_file->close;
    unlink($tmp_file) if (-e $tmp_file);

    return 0;#seb debug pour xsl inline
}

=pod

=item function IsiANTDBEntry

 Title        : IsiANTDBEntry
 Usage        : IsiANTDBEntry($input)
 Function     : Test if $input is a DB::Entry Xml file 
			 (test root node name and try to build a iANT::DB::Entry object) 
 Returns      : boolean
 Args         : $input
                $no_features : if defined, returns a light version of iANT::DB::Entry object (without any entry feature element
 Globals      : none
 
=cut

sub IsiANTDBEntry
{
    my $input = shift();
    my $no_features = shift();
    
    eval 'use iANT::DB::Entry';
    return 0 if ($@);
    my $o_db_entry;
    
my $tmp_file = "/tmp/$$" .$^T;

    if (($input =~ /\n/) || (! -e $input))
    {
    	
    	my $o_file = new IO::File(">$tmp_file");
	print $o_file $input;
	$o_file->close;
	$input = $tmp_file;
    }


    my $o_file = new IO::File($input);
    
    if (!defined $o_file)
    {
     unlink($tmp_file) if (-e $tmp_file);
     return 0;
    }
    while (my $line = <$o_file>)
    {

        chomp $line;
#        if ($line =~ /^<\?.+\?>\s*$/)#seb debug pour xsl inline
#        {#seb debug pour xsl inline
#            next;#seb debug pour xsl inline
#        }#seb debug pour xsl inline
#        else#seb debug pour xsl inline
#        {#seb debug pour xsl inline
            if ($line =~ /<entry_db\s*/)
            {
                eval {$o_db_entry = New iANT::DB::Entry($input,undef,$no_features)};            

                if ($@)
                {
                    print STDOUT "err $@";
                    $o_file->close;
                    unlink($tmp_file) if (-e $tmp_file);
                    return 0;
                }
                else
                {
                    $o_file->close;
                    unlink($tmp_file) if (-e $tmp_file);    
                    return $o_db_entry;
                }
            }
            else
            {
#                $o_file->close;#seb debug pour xsl inline
#                return 0;#seb debug pour xsl inline
                next; #seb debug pour xsl inline
            }
#        }#seb debug pour xsl inline
    }
    $o_file->close;  
    unlink($tmp_file) if (-e $tmp_file);
    return 0;#seb debug pour xsl inline
}

=pod

=item function IsiANTDBFeature

 Title        : IsiANTDBFeature
 Usage        : IsiANTDBFeature($input)
 Function     : Test if $input is a DB::Feature Xml file 
			 (test root node name and try to build a iANT::DB::Feature object) 
 Returns      : boolean
 Args         : $input
 Globals      : none
 
=cut

sub IsiANTDBFeature
{
    my $input = shift();
    eval 'use iANT::DB::Feature';
    return 0 if ($@);
    my $o_db_feature;
    	my $tmp_file = "/tmp/$$" .$^T;   
if (! -e $input)
    {

    	my $o_file = new IO::File(">$tmp_file");
	print $o_file $input;
	$o_file->close;
	$input = $tmp_file;
    }

    my $o_file = new IO::File($input);
    if (!defined $o_file)
    {
     unlink($tmp_file) if (-e $tmp_file);
     return 0;
    }

    while (my $line = <$o_file>)
    {
        chomp $line;
        if ($line =~ /^<\?.+\?>\s*$/)
        {
            next;
        }
        else
        {
            if ($line =~ /<entry_features\s?/)
            {
                eval {$o_db_feature = New iANT::DB::Feature($input)};
                if ($@)
                {unlink($tmp_file) if (-e $tmp_file);
                    return 0;
                }
                else
                {unlink($tmp_file) if (-e $tmp_file);
                    return $o_db_feature;
                }
            }
            else
            {unlink($tmp_file) if (-e $tmp_file);
                return 0;
            }
        }
    }
    unlink($tmp_file) if (-e $tmp_file);
    $o_file->close;
}

=pod

=head1 DEPRECATED 

sub SplitDBEntryXmlToEntriesXml
{
    my $db_entry_xml = shift();
    
    
    my @a_entry_xml = ();
    $db_entry_xml =~ s/^<\?.+\?>\s*\n$//;
    $db_entry_xml =~ s/<entry_db[^>]*>//;
    $db_entry_xml =~ s/<\/entry_db>//;

    my $previous_nl = $/;
    $/ = '</entry>';
    
    my @a_entry_xml_raw = split (/<\/entry>/,$db_entry_xml);
    foreach my $entry_xml_raw ( @a_entry_xml_raw)
    {
        if ($entry_xml_raw =~ /<entry /)
        {
            $entry_xml_raw .= '</entry>';
            push (@a_entry_xml,$entry_xml_raw );
        }
    }
    $/ = $previous_nl ;
    return \@a_entry_xml;
}

sub SplitDBFeatureXmlToFeaturesXml
{
    my $db_feature_xml = shift();
    my @a_feature_xml = ();
    $db_feature_xml =~ s/^<\?.+\?>\s*\n$//;
    $db_feature_xml =~ s/<entry_features[^>]*>//;
    $db_feature_xml =~ s/<\/entry_features>//;

    my $previous_nl = $/;
    $/ = '</feature>';
    
    my @a_feature_xml_raw = split (/<\/feature>/,$db_feature_xml);
    foreach my $feature_xml_raw ( @a_feature_xml_raw)
    {
        if ($feature_xml_raw =~ /<feature /)
        {
            $feature_xml_raw .= '</feature>';
            push (@a_feature_xml,$feature_xml_raw );
        }
    }
    $/ = $previous_nl ;
    return \@a_feature_xml;
}

=cut

=pod

=head3 function SplitDBFeatureXMLIntoFeaturesXML

 Title        : SplitDBFeatureXMLIntoFeaturesXML
 Usage        : SplitDBFeatureXMLIntoFeaturesXML($db_feature_xml_file)
 Function     : returns an array ref of atomic feature xml data
 Returns      : array ref
 Args         : $db_feature_xml_file : iANT::DB::Feature XML file
 Globals      : none

=cut


sub SplitDBFeatureXMLIntoFeaturesXML
{
    my $file = shift;
    
    my @a_xml_features = ();
    if (($file !~ /\n/) and (-e $file))
    {
        my $cmd = 'sed -i \'s/></>\n</g\'';
        system("$cmd $file");
        my $fh_in = new IO::File($file);

        my $feature_xml = '';
        my $in_feature = 0;
        while (my $line = <$fh_in>)
        {
            chomp $line;
            if ($line =~ /<feature /)
            {
                $feature_xml = "$line\n";
                $in_feature = 1;
            }
            elsif ($in_feature == 1)
            {
                $feature_xml .= "$line\n";
            }
            if ($line =~ /<\/feature>/)
            {
                push(@a_xml_features,$feature_xml);                
                $in_feature = 0;

            }
        }
        $fh_in->close;
    
    
    
    }
    else
    {
        $file =~ s/></>\n</g;
        my $feature_xml = '';
        my $in_feature = 0;
        my @a_lines = split ("\n",$file);
        chomp @a_lines;
        foreach my $line (@a_lines)
        {
            
            if ($line =~ /<feature /)
            {
                $feature_xml = "$line\n";
                $in_feature = 1;
            }
            elsif ($in_feature == 1)
            {
                $feature_xml .= "$line\n";
            }
            if ($line =~ /<\/feature>/)
            {
                push(@a_xml_features,$feature_xml);
                $in_feature = 0;
                
            }
        }
    }
    
    
    
    return \@a_xml_features;
}


=head3 function SplitDBEntriesXMLIntoEntriesXML

 Title        : SplitDBEntriesXMLIntoEntriesXML
 Usage        : SplitDBEntriesXMLIntoEntriesXML($db_entry_xml_file)
 Function     : returns an array ref of atomic entry xml data
 Returns      : array ref
 Args         : $db_feature_xml_file : iANT::DB::Entry XML file
 Globals      : none

=cut


sub SplitDBEntriesXMLIntoEntriesXML
{
    my $file = shift;
    
    my @a_xml_entries = ();    
    if (($file !~ /\n/) and (-e $file))
    {

        my $cmd = 'sed -i \'s/></>\n</g\'';
        system("$cmd $file");

        my $fh_in = new IO::File($file);

        my $entry_xml = '';
        my $in_entry = 0;
        while (my $line = <$fh_in>)
        {
            chomp $line;
            if ($line =~ /<entry /)
            {
                $entry_xml = "$line\n";
                $in_entry = 1;
            }
            elsif ($in_entry == 1)
            {
                $entry_xml .= "$line\n";
            }
            if ($line =~ /<\/entry>/)
            {
                push(@a_xml_entries,$entry_xml);

            }
        }
        $fh_in->close;
    }
    else
    {
        $file =~ s/></>\n</g;
        #print STDOUT "<pre>" . HTML::Entities::encode($file) . "</pre>";
        my $entry_xml = '';
        my $in_entry = 0;
        my @a_lines = split ("\n",$file);
        chomp @a_lines;
        foreach my $line (@a_lines)
        {
            next if (($line eq '') or ($line =~ /entry_db/));

            if ($line =~ /<entry /)
            {
                $entry_xml = "$line\n";
                $in_entry = 1;
            }
            elsif ($in_entry == 1)
            {
                $entry_xml .= "$line\n";
            }
            if ($line =~ /<\/entry>/)
            {

                push(@a_xml_entries,$entry_xml);
            }
        }
    }
    return \@a_xml_entries;
}


sub DeleteEntryFeaturesFromXml
{
    my $entry_xml = shift();
    $entry_xml =~ s/\n/__nl__/g;
    $entry_xml =~ s/<entry_features.+<\/entry_features>//;
    $entry_xml =~ s/__nl__/\n/g;
    return $entry_xml;
}


=pod

=item function GetDBFeatureFile

 Title        : GetDBFeatureFile
 Usage        : GetDBFeatureFile($input, $o_param)
 Function     : Look for a DB::Feature xml file from a Entry or DB::Entry file name
 Returns      : $entry_db_feature_file file name if found
			 else undef
 Args         : $input (an entry file name, an entry db file name)
			 $o_param
 Globals      : [conf] extension_iant_entry, extension_iant_db_entry, extension_iant_db_feature
 
=cut

sub GetDBFeatureFile
{
    my ($input, $o_param) = (@_);
    my $entry_db_feature_file           = $input;
    my $entry_file_extension            = $o_param->Get('extension_iant_entry');
    my $entry_db_file_extension         = $o_param->Get('extension_iant_db_entry');
    my $entry_db_feature_file_extension = $o_param->Get('extension_iant_db_feature');

    $entry_db_feature_file =~ s/$entry_file_extension$//;
    $entry_db_feature_file =~ s/$entry_db_file_extension$//;
    $entry_db_feature_file .= $entry_db_feature_file_extension;

    if (-e $entry_db_feature_file)
    {
        return $entry_db_feature_file;
    }
    return;

}

=pod

=item function BuildHtmlSelect

 Title        : BuildHtmlSelect
 Usage        : BuildHtmlSelect($data, $input_name, $select)
 Function     : Build an HTML popup menu from different kind of source:
				1. a file: value|label\nvalue|label\n...
				2. an array: (value|label,value|label...)
				3. a hash: (value => lable, value =>label, ...)
				4. a string: value|label#value|label#...
 Returns      : $popup_menu html code
 Args         : $data
			 $input_name name of html input
			 $select default selected item
 Globals      : none
 
=cut

sub BuildHtmlSelect
{
    my ($data, $input_name, $select) = (@_);

    my %h_select_data;

    if (defined $data && ($data ne ''))
    {
        if (-e $data)
        {

            my $o_file = new IO::File;
            if ($o_file->open("< $data"))
            {
                while (my $line = <$o_file>)
                {
                    chomp $line;
                    my ($value, $label) = split(/\|/, $line, 2);
                    if (!defined $label)
                    {
                        $label = $value;
                    }
                    $h_select_data{$value} = '['. $value .']' . $label;
                }
                $o_file->close;
            }
        }
        elsif (ref($data) =~ /array/i)
        {
            foreach my $item (@{$data})
            {
                chomp $item;
                my ($value, $label) = split(/\|/, $item, 2);
                if (!defined $label)
                {
                    $label = $value;
                }
                $h_select_data{$value} = '['. $value .']' . $label;
            }
        }
        elsif (ref($data) =~ /hash/i)
        {
            %h_select_data = %{$data};
        }
        else
        {
            #my @a_fields = split(/#/, $data);
            #le # gene le parsing par AnotherTemplate
            my @a_fields = split(/,/, $data);
            
            foreach my $item (@a_fields)
            {
                chomp $item;
                my ($value, $label) = split(/\|/, $item, 2);
                if (!defined $label)
                {
                    $label = $value;
                }
                $h_select_data{$value} = '['. $value .']' . $label;
            }
        }

        my $o_cgi = New CGI();
        my $popup_menu = $o_cgi->popup_menu(
                                            -name    => $input_name,
                                            -default => $select,
                                            -values  => [sort (keys %h_select_data)],
                                            -labels  => \%h_select_data
                                            );

        return $popup_menu;
    }
    else
    {
        return;
    }

}

=pod

=item function GetValue

 Title        : GetValue
 Usage        : GetValue($data, $select)
 Function     : Get label associated to a value from different kind of source 
				1. a file: value|label\nvalue|label\n...
				2. an array: (value|label,value|label...)
				3. a hash: (value => lable, value =>label, ...)
				4. a string: value|label#value|label#...
 Returns      : $label
 Args         : $data
			 $select selected item
 Globals      : none
 
=cut

sub GetValue
{
    my ($data, $select) = (@_);

    my %h_select_data;

    if (defined $data && ($data ne ''))
    {
        if (-e $data)
        {
            my $o_file = new IO::File;
            if ($o_file->open("< $data"))
            {
                while (my $line = <$o_file>)
                {
                    chomp $line;
                    my ($value, $label) = split(/\|/, $line, 2);
                    if (!defined $label)
                    {
                        $label = $value;
                    }
                    $h_select_data{$value} = $label;
                }
                $o_file->close;
            }
        }
        elsif (ref($data) =~ /array/i)
        {
            foreach my $item (@{$data})
            {
                chomp $item;
                my ($value, $label) = split(/\|/, $item, 2);
                if (!defined $label)
                {
                    $label = $value;
                }
                $h_select_data{$value} = $label;
            }
        }
        elsif (ref($data) =~ /hash/i)
        {
            %h_select_data = %{$data};
        }
        else
        {
            my @a_fields = split(/#/, $data);
            foreach my $item (@a_fields)
            {
                chomp $item;
                my ($value, $label) = split(/\|/, $item, 2);
                if (!defined $label)
                {
                    $label = $value;
                }
                $h_select_data{$value} = $label;
            }
        }

        return $h_select_data{$select};
    }
    else
    {
        return;
    }

}



sub BuildFastaHeader
{
	
    my $param_entry = shift();
    my $add_locus_to_id = shift;
    my $o_param = shift();
    
    
    $o_param = New ParamParser('INIT') if (! defined $o_param);
    my $id        = '';
    my $accession = '';    
    my $locus_tag = '';
    my $begin = '';
    my $end = '';
    my $strand = '';
    my $structure = '';
    my $len       = '';
    my $species   = '';
    my $type      = '';
    my $def       = '';
    my $pm        = '';
    my $ec        = '';
    my $synonyms  = '';
    my $chr       = '';
    my $dbsource    = '';
    my $ipr_def       = '';
    
    if (ref($param_entry) =~ /entry/i)
    {
        my $o_entry = $param_entry;
        $id        = $o_entry->Get('id');	
        if ($id eq '')
        {
  	        $id = $o_entry->Get('locus_tag') 
        }
        else
        {
            my $locus_tag =  $o_entry->Get('locus_tag');
            $id .= '-' .  $locus_tag if (defined $add_locus_to_id && ($locus_tag ne '') && $id !~ /-$locus_tag/);
        }
        $accession = $o_entry->Get('accession');
        $locus_tag = $o_entry->Get('locus_tag');

        $len       = $o_entry->Get('length');
        $species   = $o_entry->GetOrganism->GetSpecies->GetCode();
        $type      = $o_entry->Get('type');

        my $ra_o_definition = $o_entry->GetDefinitions();
        my @a_pmid          = ();
        my @a_ecnumbers          = ();

		my @a_synonyms = ();
        foreach my $o_definition (@{$ra_o_definition})
        {
            if ($o_definition->GetOntology() eq 'pubmed')
            {
                push(@a_pmid, $o_definition->GetId());
            }
            elsif ($o_definition->GetOntology() =~ /enzyme/i)
            {
                push(@a_ecnumbers, $o_definition->GetId());
            }
            elsif (($o_definition->GetOntology() eq 'NONE') && ($o_definition->GetId() eq 'product'))
            {
                $def = $o_definition->GetContent();
                $def =~ s/^\s+//;
            }
            elsif (($o_definition->GetOntology() eq 'INTERPRO') && ($o_definition->GetId() eq 'product'))
            {
                $ipr_def = $o_definition->GetContent();
                $ipr_def =~ s/^\s+//;
            }
            elsif ($o_definition->GetOntology() eq 'SYNONYM')
            {
                push(@a_synonyms, $o_definition->GetId());
            }
        }
        $pm = join(';', @a_pmid);
        $synonyms = join(';', @a_synonyms);
        $ec = join(';', @a_ecnumbers);

        my $ra_o_link = $o_entry->GetLinks();
#        my @a_contigs = ();

        foreach my $o_link (@{$ra_o_link})
        {
            if ($o_link->GetType() eq 'partOf')
            {
                $chr = $o_link->GetAccession();
                $begin = $o_link->GetBegin();
                $end = $o_link->GetEnd();
                $strand = $o_link->GetStrand();
                $structure = $o_link->GetStructure();
                last;
            }
            elsif ($o_link->GetType() eq 'producedFrom')
            {
                $chr = $o_link->GetAccession();
                $begin = $o_link->GetBegin();
                $end = $o_link->GetEnd();
                $strand = $o_link->GetStrand();
                $structure = $o_link->GetStructure();
                last;
            }
        }

#        $chr = join(';', @a_contigs);
    }
    elsif (ref($param_entry) =~ /hash/i)
    {
        my $rh_sequence_info = $param_entry;
        $id        = $rh_sequence_info->{'id'}        if (defined $rh_sequence_info->{'id'});
        $accession = $rh_sequence_info->{'accession'} if (defined $rh_sequence_info->{'accession'});
        $locus_tag = $rh_sequence_info->{'locus_tag'} if (defined $rh_sequence_info->{'locus_tag'});
        $len       = $rh_sequence_info->{'length'}    if (defined $rh_sequence_info->{'length'});
        $species   = $rh_sequence_info->{'species'}   if (defined $rh_sequence_info->{'species'});
        $type      = $rh_sequence_info->{'type'}      if (defined $rh_sequence_info->{'type'});
	$dbsource    = $rh_sequence_info->{'dbsource'}      if (defined $rh_sequence_info->{'dbsource'});
        my @a_pmid = ();
        my @a_ecnumbers          = ();
		my @a_synonyms          = ();

        foreach my $rh_definition (@{$rh_sequence_info->{'ra_definitions'}})
        {
            if ($rh_definition->{'ontology'} eq 'pubmed')
            {
                push(@a_pmid, $rh_definition->{'id'});
            }
            elsif ($rh_definition->{'ontology'} =~ /enzyme/i)
            {
                push(@a_ecnumbers, $rh_definition->{'id'});
            }
            elsif ($rh_definition->{'ontology'} =~ /synonym/i)
            {
                push(@a_synonyms, $rh_definition->{'id'});
            }
            
            elsif (($rh_definition->{'ontology'} eq 'NONE') && ($rh_definition->{'id'} eq 'product'))
            {
                $def = $rh_definition->{'content'};
                $def =~ s/^\s+//;
            }
        }
        $pm = join(';', @a_pmid);
        $ec = join(';', @a_ecnumbers);
		$synonyms = join(';', @a_synonyms);
		
        my @a_contigs = ();
        foreach my $rh_link (@{$rh_sequence_info->{'ra_links'}})
        {
            if ($rh_link->{'type'} eq 'partOf')
            {
                #push(@a_contigs, $rh_link->{'accession'});
                $chr =$rh_link->{'accession'} if (defined $rh_link->{'accession'});
                $begin =$rh_link->{'begin'} if (defined $rh_link->{'begin'});
                $end =$rh_link->{'end'} if (defined $rh_link->{'end'});
                $strand =$rh_link->{'strand'} if (defined $rh_link->{'strand'});
                $structure =$rh_link->{'structure'} if (defined $rh_link->{'structure'});
            }
        }
        #$chr = join(';', @a_contigs);
    }

    #>id acc=sp:P15711 len=87 chr=1 sp=code_prj k=[EBVA] pm=1689460;1689461 type=gene|cds|protein def=description

    my %h_known_database = (
                            'gb'   => 'gb|accession|locus',
                            'emb'  => 'emb|accession|locus',
                            'dbj'  => 'dbj|accession|locus',
                            'pir'  => 'pir|entry',
                            'prf'  => 'prf|name',
                            'sp'   => 'sp|accession|entry name',
                            'pdb'  => 'pdb|entry|chain',
                            'pat'  => 'pat|country|number',
                            'bbs'  => 'bbs|number',
                            'gnl'  => 'gnl|database|identifier',
                            'ref'  => 'ref|accession|locus',
                            'lcl'  => 'lcl|identifier',
			    'et'   => 'et|accession',
                            'nc'  => 'lcl|locus',
                            'none' => 'identifier',
			    'gi' => 'gi|identifier|dbsource|accession|',
                            );
    my $fasta_header = '>';
    my $id_syntax    = '';
    
    my $database     = $o_param->Get('database');
    if (!$o_param->IsDefined('database'))
    {

        #cas ou le id est deja formate
        foreach my $known_database (keys %h_known_database)
        {
            if ($id =~ /^$known_database\|/)
            {
                $id_syntax = $h_known_database{$known_database};
                last;
            }
        }
        $id_syntax = $h_known_database{'lcl'} if ($id_syntax eq '');
    }
    elsif (!defined $h_known_database{$database})
    {
        $id_syntax = $h_known_database{'lcl'};
    }
    else
    {
        $id_syntax = $h_known_database{$database};
    }


    if ($id =~ /\S+\|(\S+)$/)
    {
        $id = $1;
    }
    $id_syntax =~ s/accession/$accession/;
    $id_syntax =~ s/locus/$id/;
    $id_syntax =~ s/identifier/$id/;
    $id_syntax =~ s/entry name/$id/;
    $id_syntax =~ s/entry/$accession/;
    $id_syntax =~ s/database/$database/;
    $id_syntax =~ s/dbsource/$dbsource/;

    
    if ($type eq '' && $o_param->IsDefined('type'))
    {
        $type = $o_param->Get('type');
    }

    $o_param->Set('type', $type);

    $def =~ s/\s+$//;
    $def .= ' ' if ($def ne '');
    $def .= $ipr_def if (($def eq '') or ($o_param->IsDefined('interpro_product')));

    $fasta_header .= $id_syntax;
    $fasta_header .= " acc=$accession";    
    $fasta_header .= " locus_tag=$locus_tag" if ($locus_tag ne '');
    $fasta_header .= " len=$len";
    $fasta_header .= " sp=$species";
    $fasta_header .= " type=$type";
    $fasta_header .= " begin=$begin";
    $fasta_header .= " end=$end";
    $fasta_header .= " strand=$strand";
    $fasta_header .= " structure=$structure" if ($structure ne '');
    $fasta_header .= " chr=$chr";
    $fasta_header .= " pm=$pm" if ($pm ne '');
    $fasta_header .= " ec=$ec" if ($ec ne '');
    $fasta_header .= " synonyms=$synonyms" if ($synonyms ne '');
        
    $fasta_header .= " def=$def";

    return $fasta_header;

}

sub ParseFastaHeader
{
    my $fasta_header    = shift;
    my $database = shift;
    my %h_sequence_info = ();
    my @a_definitions   = ();
    my @a_links         = ();
    my ($id) = ($fasta_header =~ /^>?(\S+)/);
    my $accession = $id;    
    my $dbsource;
    my %h_known_database = (
                            'gb'   => '($accession,$locus_tag) = ($fasta_header =~ />?gb\|(\w+)\|(\w+)/)',
                            'emb'  => '($accession,$locus_tag) = ($fasta_header =~ />?emb\|(\w+)\|(\w+)/)',
                            'dbj'  => '($accession,$locus_tag) = ($fasta_header =~ />?dbj\|(\w+)\|(\w+)/)',
                            'sp'   => '($accession,$id) = ($fasta_header =~ />?sp\|(\w+)\|(\w+)/)',
                            'gnl'  => '($database,$id) = ($fasta_header =~ />?gnl\|(\w+)\|(\w+)/)',
                            'ref'  => '($accession,$locus_tag) = ($fasta_header =~ />?ref\|(\w+)\|(\w+)/)',
                            'lcl'  => '($id,$accession) = ($fasta_header =~ />?lcl\|(\S+)\|?(\S+)?/)',
                            'nc'  => '($locus_tag) = ($fasta_header =~ />?lcl\|(\S+)/)',
                            'none' => '($id) = ($fasta_header =~ />?(\S+)/)',
			    'gi'   => '($id,$dbsource,$accession) = ($fasta_header =~ />?gi\|(\d+)\|(emb|gb|ref)\|(.+)\|/)',
                            );
                            
    
    my $namespace =  $database;
    if (! defined $database)
    {
        ($namespace) = ($fasta_header =~ /^>?(\w+)\|/);
    }
    
    if (defined $namespace &&  defined $h_known_database{$namespace})
    {
        my $regexp = $h_known_database{$namespace};

        eval ($regexp);  
        $database = $namespace;
    	$accession = $id if (! defined $accession);
	}

    else
    {
        if ($accession =~ /\w+\|([^|]+)/)
        {
	    $accession = $1;
        }

        if (defined $database) 
        {
            if ($database eq 'sp')
            {
                ($id, $accession) = ($fasta_header =~ /^>?(\S+)\s+\((\S+)\)/);
                $fasta_header =~ s/^>?\S+\s+\(\S+\)//;
                ($h_sequence_info{species}) = ($id =~ /\w+_(\w+)/);

            }
            else
            {
                $accession = $id;
	        if ($accession =~ /\w+\|([^|]+)/)
    	        {
	            $accession = $1;
    	        }
                $fasta_header =~ s/^>?\S+//;

            }
        }
        elsif (!defined $namespace or (! defined  $h_known_database{$namespace}))
        {
            $accession = $id;
		    if ($accession =~ /\w+\|([^|]+)/)
    	    {
	        $accession = $1;
    	    }
            $fasta_header =~ s/^>?\S+//;

        }
    }
    $h_sequence_info{id}        = $id;
    $h_sequence_info{accession} = $accession;
    $h_sequence_info{dbsource}  = $dbsource;

    my $raw_definition = '';
    
    if ($fasta_header !~ /\S+=\S+/)
    {
        my ($definition) = ($fasta_header =~ /^\s+(.+)/);
        if (!defined $definition)
        {
            $definition = '';
        }
#        push(@a_definitions, {ontology => 'NONE', id => 'product', content => "$definition"});
        $raw_definition = $definition;
    }
    elsif ($fasta_header =~ / def=(.+)/)
    {
#        push(@a_definitions, {ontology => 'NONE', id => 'product', content => "$1"});
        $raw_definition = $1;        
        my $value = $1;
        $value =~ s/\|/\\|/g;
        $value =~ s/\]/\\]/g;
        $value =~ s/\[/\\[/g;
        $value =~ s/\{/\\{/g;
        $value =~ s/\}/\\}/g;
        $value =~ s/\+/\\+/g;
	$value =~ s/\)/\\)/g;# if ($value =~ /[^\\]\)/);
        $value =~ s/\(/\\(/g;# if ($value =~ /[^\\]\(/);
	$fasta_header =~ s/ def=$value//;
    }

    while ($fasta_header =~ /(\S+)=(\S+)/)
    {

        #>id acc=sp:P15711 len=87 chr=1 sp=code_prj k=[EBVA] pm=1689460;1689461 type=gene|cds|protein def=description
        $h_sequence_info{accession} = $2 if ($1 eq 'acc');
        $h_sequence_info{length}    = $2 if ($1 eq 'len');
        $h_sequence_info{type}      = $2 if ($1 eq 'type');
        $h_sequence_info{species}   = $2 if ($1 eq 'sp');
        $h_sequence_info{locus_tag}   = $2 if ($1 eq 'locus_tag');
        $h_sequence_info{version}   = $2 if ($1 eq 'version');
	$h_sequence_info{structure}   = $2 if ($1 eq 'version');

        $h_sequence_info{begin} = $2 if ($1 eq 'begin');
        $h_sequence_info{end}    = $2 if ($1 eq 'end');
        $h_sequence_info{strand}    = $2 if ($1 eq 'strand');
        $h_sequence_info{partofaccession}    = $2 if ($1 eq 'chr');

        if ($1 eq 'pm')
        {
            my @a_pmid = split(/;/, $2);
            foreach my $pmid (@a_pmid)
            {
                push(@a_definitions, {ontology => 'pubmed', id => $pmid});
            }
        }
        
        if ($1 eq 'ec')
        {
            my @a_ecnumbers = split(/;/, $2);
            foreach my $ec (@a_ecnumbers)
            {
                push(@a_definitions, {ontology => 'ENZYME', id => $ec});
            }
        }
        
		if ($1 eq 'synonyms')
        {
            my @a_synonyms = split(/;/, $2);
            foreach my $synonym (@a_synonyms)
            {
                push(@a_synonyms, {ontology => 'SYNONYM', id => $synonym});
            }
        }
       
        
        $h_sequence_info{organelle} = "$2" if ($1 eq 'chr');
        my $key = $1;
        my $value = $2;
        #my @a_spec_char = ('[',']','(',')','{','}','|','\\','+');
        #foreach my $spec_char (@a_spec_char)
        #{
        #    $value =~ s/\$spec_char/\\\$spec_char/g;
        #}
        
        $value =~ s/\|/\\|/g;
        $value =~ s/\]/\\]/g;
        $value =~ s/\[/\\[/g;
        $value =~ s/\{/\\{/g;
        $value =~ s/\}/\\}/g;
        $value =~ s/\+/\\+/g;        
	
	$value =~ s/\)/\\)/g   if ($value =~ /[^\\]\)/);
        $value =~ s/\(/\\(/g  if ($value =~ /[^\\]\(/);

        #$fasta_header =~ s/\(/\\(/g  if ($fasta_header =~ /[^\\]\)/);
        #$fasta_header =~ s/\)/\\)/g  if ($fasta_header =~ /[^\\]\(/);
        
        $fasta_header =~ s/$key=$value//g;
    }
    
    push(@a_links, {type => 'partOf', accession => $h_sequence_info{partofaccession}, begin=>$h_sequence_info{begin}, end=>$h_sequence_info{end},strand=>$h_sequence_info{strand} }) if (defined $h_sequence_info{partofaccession});

    $fasta_header =~ s/^>?$id// if (defined $id);
    if ($fasta_header !~ /^\s+$/)
    {
        $fasta_header =~ s/\s+$//;
        $fasta_header =~ s/^\s+//;
        $raw_definition .= ' ' . $fasta_header;
    }
    
    $raw_definition =~ s/^\s+//;
    $raw_definition = &__CleanRawDefinition($raw_definition);
    push(@a_definitions, {ontology => 'NONE', id => 'product', content => "$raw_definition"});
    $h_sequence_info{raw_definition} = $raw_definition;
    $h_sequence_info{ra_definitions} = \@a_definitions;
    $h_sequence_info{ra_links}       = \@a_links;

    
    return \%h_sequence_info;
}

sub __CleanRawDefinition
{
    my $raw_definition = shift;
    
    $raw_definition =~ s/>\S+//;
    $raw_definition =~ s/ def=(\S+)/ def= $1/g;
    $raw_definition =~ s/\w+=\s//g;
    $raw_definition =~ s/\w+=\S+//g;
    $raw_definition =~ s/\s+/ /;
    $raw_definition =~ s/^\s//;
    $raw_definition =~ s/\s+$//g;
    $raw_definition =~ s/\w+=$//g;
    
    
    return $raw_definition;
}

sub BuildEntryObject
{
    my ($rh_sequence_info, $sequence) = (@_);

    my $o_entry = New iANT::Entry();


    if (defined $rh_sequence_info && (ref($rh_sequence_info) =~ /hash/i))
    {
        if (ref($rh_sequence_info->{ra_definitions}) =~ /array/i)
        {
            foreach my $rh_definition (@{$rh_sequence_info->{ra_definitions}})
            {
                my $o_entry_definition = New iANT::Entry::Definition($rh_definition);
                $o_entry->SetDefinition($o_entry_definition);
            }
            delete $rh_sequence_info->{ra_definitions};
        }

        if (ref($rh_sequence_info->{ra_links}) =~ /array/i)
        {
            foreach my $rh_link (@{$rh_sequence_info->{ra_links}})
            {
                my $o_entry_link = New iANT::Entry::Link($rh_link);
                $o_entry->SetLink($o_entry_link);
            }
            delete $rh_sequence_info->{ra_links};
        }
        $rh_sequence_info->{species} = '' if (!defined $rh_sequence_info->{species});
        my $o_organism_species = New iANT::Organism::Species({code => $rh_sequence_info->{species}});
        delete $rh_sequence_info->{species};

        my $o_organism_taxonomy = New iANT::Organism::Taxonomy();
        $rh_sequence_info->{organelle} = '' if (!defined $rh_sequence_info->{organelle});
        delete $rh_sequence_info->{organelle};
        my $o_organism_organelle = New iANT::Organism::Organelle({content => $rh_sequence_info->{organelle}});

        my $o_organism =
          New iANT::Organism(
                             {
                              o_species   => $o_organism_species,
                              o_taxonomy  => $o_organism_taxonomy,
                              o_organelle => $o_organism_organelle
                             }
                             );

        $o_entry->SetOrganism($o_organism);
        my $o_entry_sequence = New iANT::Entry::Sequence({content => $sequence});
        $o_entry->SetSequence($o_entry_sequence);
        $o_entry->Update($rh_sequence_info);

        #print STDERR "species " . $o_entry->GetOrganism . "\n";exit;

        return $o_entry;
    }
}

sub FormatSequenceLine
{
    my ($sequence, $length_line) = (@_);

    $sequence =~ s/\?/X/g;
    $sequence =~ s/\W//g;
    #if (&__ControlSequenceType($sequence))
    #{
#    $sequence =~ s/\s//g;

        my ($size_sequence) = length($sequence);
        my ($nb_lines)      = $size_sequence / $length_line;
        my ($j)             = 0;
        $nb_lines = int($nb_lines) + 1 if ($nb_lines =~ /\./);
        my $formatted_sequence = '';
        for ($j = 0; $j < $nb_lines; $j++)
        {
            $formatted_sequence .= substr($sequence, ($j) * $length_line, $length_line) . "\n";
        }
        return $formatted_sequence;
    #}

}

sub GetEnzymeInfo
{
    my ($ec_code) = (@_);
    use ParamParser;
    my $o_param;
    $o_param = &iANT::Cfg::Load($o_param);

    my $enzyme_template = $o_param->Get('tpl_dir') . '/enzyme.xhtml';

    use AnotherTemplate;
    my $o_AnotherTemplate = New AnotherTemplate();
    $o_AnotherTemplate->Update($enzyme_template);
    $o_AnotherTemplate->AddVariable('ec_number',$ec_code);
    $o_AnotherTemplate->Parse();
    my $parsed_template = $o_AnotherTemplate->Get('parsed_template');

    return $parsed_template;

}

sub GetGODescription
{
    my ($go_term) = (@_);
    use ParamParser;
    my $o_param;
    $o_param = &iANT::Cfg::Load($o_param);
    
    my $moby_uri = "http://moby.toulouse.inra.fr/MOBY/Central";
    my $moby_url = "http://moby.toulouse.inra.fr/cgi/MOBY-Central.pl";

    #my $go_description_tpl = "#BIOMOBY|MOBYSHoundGODBGetNameByID\@bioinfo.icapture.ubc.ca\@$moby_uri,$moby_url|inputs => [{namespace => 'GO', name => 'go_id', type => 'Object', id => '$go_term', content => ''}], output => 'raw'#";
    #use AnotherTemplate;
    #my $o_AnotherTemplate = New AnotherTemplate();
    #$o_AnotherTemplate->Update($go_description_tpl);
    #$o_AnotherTemplate->Parse();
    #my $go_description= $o_AnotherTemplate->Get('parsed_template');
    my $go_description = 'unavailable';
    return $go_description;

}

sub GetGOClassification
{
    my ($go_term) = (@_);
    use ParamParser;
    my $o_param;
    $o_param = &iANT::Cfg::Load($o_param);
    
    my $moby_uri = "http://moby.toulouse.inra.fr/MOBY/Central";
    my $moby_url = "http://moby.toulouse.inra.fr/cgi/MOBY-Central.pl";
	
	#my $moby_uri = "http://moby.ucalgary.ca/MOBY/Central";
    #my $moby_url =  "http://moby.ucalgary.ca/moby/MOBY-Central.pl";
	
	my $go_classification = 'unavailable';
    
	if (0)
	{
		my $go_classification_tpl= "#BIOMOBY|MOBYSHoundGODBGetClassification\@bioinfo.icapture.ubc.ca\@$moby_uri,$moby_url|inputs => [{namespace => 'GO', name => 'go_id', type => 'Object', id => '$go_term', content => ''}], output => 'raw'#";
		use AnotherTemplate;
		my $o_AnotherTemplate = New AnotherTemplate();
		$o_AnotherTemplate->Update($go_classification_tpl);
		$o_AnotherTemplate->Parse();
		$go_classification= $o_AnotherTemplate->Get('parsed_template');
    }
    
    return $go_classification;

}



1;
